"""Interfacies for function communication"""
from dataclasses import dataclass
from typing import Dict, Mapping, Optional, Union
import numpy as np
from numpy.typing import NDArray

TransformationMatrixDict = Mapping[
    str, Union[Dict[str, NDArray[np.float64]], NDArray[np.float64]]
]


@dataclass
class WriterInput:
    """Inputs for the writer classes"""

    output: str
    transformation_matrices: TransformationMatrixDict
    size: float
    show_beam: bool = False
    config_dict: Optional[Dict[str, Dict[str, str]]] = None
    shape: str = "cone"



"""Functions for creating a gltf cube mesh file"""
from sys import version_info
import logging
from typing import Any, Dict, List, Mapping, Union
import numpy as np
from numpy.typing import NDArray
import pygltflib

from nexus3d.formats.mesh import (
    create_cube_arrays,
    create_cone_arrays,
    get_mesh_from_stl,
)
from nexus3d.formats.interfaces import WriterInput
from nexus3d.matrix import rotate, translate

TransformationMatrix = Union[Dict[str, NDArray[np.float64]], NDArray[np.float64]]
TransformationMatrixDict = Mapping[str, TransformationMatrix]

logger = logging.getLogger(__name__)


def get_binary_blobs(indices: NDArray[np.uint8], vertices: NDArray[np.float32]):
    """Converts points and triangles arrays to binary blobs.

    Args:
        vertices (NDArray[np.float32]): The vertices array
        indices (NDArray[np.uint8]): The indices array

    Returns:
        (bytes, bytes): The vertices and indices binary blob
    """
    vertices_bin = vertices.flatten().tobytes()
    indices_bin = indices.tobytes()

    return indices_bin, vertices_bin


def set_data(
    gltf: pygltflib.GLTF2,
    indices_list: List[NDArray[np.uint8]],
    vertices_list: List[NDArray[np.float32]],
):
    """Writes all binary data related structures into the gltf object.
    This includes accessors, bufferViews and buffers.
    The accessors indices are the same as in the respective indices and vertices list.
    The indices_list and vertices_list must have the same length.

    Args:
        gltf (pygltflib.GLTF2): The gltf object into which to write the data.
        indices_list (List[NDArray[np.uint8]]): List containing arrays of indices.
        vertices_list (List[NDArray[np.float32]]): List containting array of vertices.
    """

    if len(indices_list) != len(vertices_list):
        raise ValueError("Indices list and vertices list must have the same length.")

    gltf_type = {
        "int8": pygltflib.BYTE,
        "uint8": pygltflib.UNSIGNED_BYTE,
        "int16": pygltflib.SHORT,
        "uint16": pygltflib.UNSIGNED_SHORT,
        "uint32": pygltflib.UNSIGNED_INT,
        "float32": pygltflib.FLOAT,
    }

    binary_data = b""
    offset = 0
    gltf.accesors = []
    gltf.bufferViews = []

    for i, (indices, vertices) in enumerate(zip(indices_list, vertices_list)):
        indices_bin, vertices_bin = get_binary_blobs(indices, vertices)

        gltf.accessors.append(
            pygltflib.Accessor(
                bufferView=2 * i,
                componentType=gltf_type[str(indices.dtype)],
                count=indices.size,
                type=pygltflib.SCALAR,
                max=[int(indices.max())],
                min=[int(indices.min())],
            )
        )

        gltf.accessors.append(
            pygltflib.Accessor(
                bufferView=2 * i + 1,
                componentType=gltf_type[str(vertices.dtype)],
                count=len(vertices),
                type=pygltflib.VEC3,
                max=vertices.max(axis=0).tolist(),
                min=vertices.min(axis=0).tolist(),
            )
        )

        gltf.bufferViews.append(
            pygltflib.BufferView(
                buffer=0,
                byteOffset=offset,
                byteLength=len(indices_bin),
                target=pygltflib.ELEMENT_ARRAY_BUFFER,
            )
        )

        fill_offset = offset + len(indices_bin) % len(vertices_bin)
        fill_offset_bin = b"\x00" * fill_offset

        gltf.bufferViews.append(
            pygltflib.BufferView(
                buffer=0,
                byteOffset=offset + len(indices_bin) + fill_offset,
                byteLength=len(vertices_bin),
                target=pygltflib.ARRAY_BUFFER,
            )
        )

        offset += len(indices_bin) + len(vertices_bin) + fill_offset
        binary_data += indices_bin + fill_offset_bin + vertices_bin

    gltf.buffers = [pygltflib.Buffer(byteLength=offset)]
    gltf.set_binary_blob(binary_data)


def clean_name(name: str, entry_name: str):
    """Cleans the name of a matrix path.

    Args:
        name (str): The name of the transformation path
        entry_name (str): The name of the current entry
    """

    def remove_prefix(text, prefix):
        if text.startswith(prefix):
            return text[len(prefix) :]
        return text

    if version_info.minor < 9:
        return remove_prefix(
            remove_prefix(remove_prefix(name, "/entry/"), entry_name),
            "/transformations/",
        ).replace("/transformations", "")

    return (
        name.removeprefix("/entry/")
        .removeprefix(entry_name)
        .removeprefix("/transformations/")
        .replace("/transformations", "")
    )


def apply_stl_transformations(
    config_dict: Dict[str, Any], matrix: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Applies stl transformations from a config dict at the end of
    a transformation matrix.

    Args:
        config_dict (Dict[str, Any]): The config dict to read transformations from.
        matrix (NDArray[np.float64]): The matrix to which append the transformations.

    Returns:
        NDArray[np.float64]: The transformed matrix
    """
    translations = np.zeros(3)
    rot_matrix = np.identity(4)
    apply_translation = apply_rotation = False

    for i, axis in enumerate(["x", "y", "z"]):
        translations[i] = config_dict.get(axis, 0)
        apply_translation = True

    for rot, rot_axis in zip(
        ["rot_x", "rot_y", "rot_z"],
        [np.array([1, 0, 0]), np.array([0, 1, 0]), np.array([0, 0, 1])],
    ):
        if rot in config_dict:
            rot_matrix = rotate(np.deg2rad(config_dict[rot]), rot_axis) @ rot_matrix
            apply_rotation = True

    if apply_translation:
        matrix = matrix @ translate(translations)

    if apply_rotation:
        matrix = matrix @ rot_matrix

    return matrix


def write_gltf_file(cli_input: WriterInput):
    """Writes a cube mesh from the transformation matrices to a gltf file.

    Args:
        filename (str): The filename to write to.
        transformation_matrices (TransformationMatrixDict): The transformation matrix dict.
        scale (float, optional): The scale of the cubes. Defaults to 0.1.
        show_beam (bool, optional): Whether to show the beam in the gltf file. Defaults to True.
    """

    def add_stl_shift(name: str, matrix: TransformationMatrix) -> TransformationMatrix:
        if name not in cli_input.config_dict:
            return matrix

        if isinstance(matrix, dict):
            matrix["stl_shift"] = apply_stl_transformations(
                cli_input.config_dict[name], matrix[next(reversed(matrix))]
            )
            return matrix

        return apply_stl_transformations(cli_input.config_dict[name], matrix)

    def append_nodes(mesh_indices: Dict[str, int]):
        for name, matrix in cli_input.transformation_matrices.items():
            children = []
            if isinstance(matrix, dict):
                for j, (cname, cmat) in enumerate(
                    add_stl_shift(name, matrix).items()  # type: ignore
                ):
                    gltf.nodes.append(
                        pygltflib.Node(
                            mesh=mesh_indices[name],
                            matrix=list(cmat.T.flat),
                            name=f"{j}-{clean_name(cname, name)}",
                        )
                    )
                    children.append(len(gltf.nodes) - 1)

            gltf.nodes.append(pygltflib.Node(mesh=mesh_indices[name], name=name))

            if children:
                gltf.nodes[-1].children = children
            else:
                gltf.nodes[-1].matrix = list(
                    add_stl_shift(name, matrix).T.flat  # type: ignore
                )

            gltf.scenes[gltf.scene].nodes.append(len(gltf.nodes) - 1)

    def append_mesh(mode: int = 4):
        gltf.meshes.append(
            pygltflib.Mesh(
                primitives=[
                    pygltflib.Primitive(
                        attributes=pygltflib.Attributes(POSITION=2 * len(indices) - 1),
                        indices=2 * len(indices) - 2,
                        mode=mode,
                    )
                ]
            ),
        )

    def create_meshs():
        shapes = {"cube": create_cube_arrays, "cone": create_cone_arrays}
        mesh_indices = {}
        shape_index = None
        for name in cli_input.transformation_matrices:
            if name in cli_input.config_dict and "file" in cli_input.config_dict[name]:
                stl_indices, stl_vertices = get_mesh_from_stl(
                    cli_input.config_dict[name]["file"],
                    cli_input.config_dict[name].get("unit"),
                )

                indices.append(stl_indices)
                vertices.append(stl_vertices)
                mesh_indices[name] = len(vertices) - 1
                append_mesh()
                continue

            if shape_index is None:
                if cli_input.shape not in shapes:
                    logger.warning(
                        "Shape `%s` not valid. Using cones as default.", cli_input.shape
                    )
                shape_indices, shape_vertices = shapes.get(
                    cli_input.shape, create_cone_arrays
                )(cli_input.size / 2)
                indices.append(shape_indices)
                vertices.append(shape_vertices)
                shape_index = len(vertices) - 1
                append_mesh()
            mesh_indices[name] = shape_index

        return mesh_indices

    if cli_input.config_dict is None:
        cli_input.config_dict = {}

    gltf = pygltflib.GLTF2(
        scene=0,
        scenes=[pygltflib.Scene(nodes=[])],
    )

    indices = []
    vertices = []
    mesh_indices = create_meshs()

    append_nodes(mesh_indices)

    if cli_input.show_beam:
        vertices.append(np.array([[0, 0, 0], [0, 0, -1]], dtype="float32"))
        indices.append(np.array([[0, 1]], dtype="uint8"))
        gltf.nodes.append(pygltflib.Node(mesh=len(indices) - 1, name="beam"))

        gltf.scenes[gltf.scene].nodes.append(len(gltf.nodes) - 1)

        append_mesh(mode=1)

    set_data(gltf, indices, vertices)

    gltf.save(cli_input.output)



"""Creates test files for rotation and translation for reference."""
import numpy as np
from nexus3d.formats.gltf_writer import write_gltf_file
from nexus3d.formats.interfaces import WriterInput
from nexus3d.matrix import rotate, translate


def create_rot_test_file(angle: float = 20, gltf: bool = False):
    """Creates a rotation test file containing rotation around all three coordinate axes."""
    rot_vecs = {
        f"rot_x_{angle}": np.array([1, 0, 0]),
        f"rot_y_{angle}": np.array([0, 1, 0]),
        f"rot_z_{angle}": np.array([0, 0, 1]),
    }

    rotation_matrices = {}
    for name, rot_vec in rot_vecs.items():
        rotation_matrices[name] = rotate(np.deg2rad(angle), rot_vec)

    write_gltf_file(
        WriterInput(
            output="rot.gltf" if gltf else "rot.glb",
            transformation_matrices=rotation_matrices,
            size=0.1,
        )
    )


def create_trans_test_file(distance: float = 0.1, gltf: bool = False):
    """Creates a translation test file containing translations along all three coordinate axes."""
    trans_matrices = {
        f"trans_x_{distance}": translate(np.array([distance, 0, 0])),
        f"trans_y_{distance}": translate(np.array([0, distance, 0])),
        f"trans_z_{distance}": translate(np.array([0, 0, distance])),
    }

    write_gltf_file(
        WriterInput(
            output="trans.gltf" if gltf else "trans.glb",
            transformation_matrices=trans_matrices,
            size=0.1,
        )
    )


def create_test_files(gltf: bool = False):
    """Creates a translation and a rotation test file."""
    create_trans_test_file(gltf=gltf)
    create_rot_test_file(gltf=gltf)



"""Create a coordiante system from NXtransformation groups"""
from dataclasses import dataclass
from functools import partial
from typing import Dict
import numpy as np
from numpy.typing import NDArray

from nexus3d.nexus_transformations import transformation_matrices_from


@dataclass
class CoordinateSystem:
    """Represents a 3D coordinate system with its origin and axes."""

    origin: NDArray[np.float64]
    x_axis: NDArray[np.float64]
    y_axis: NDArray[np.float64]
    z_axis: NDArray[np.float64]


def coord_systems_from(
    fname: str, include_process: bool = False
) -> Dict[str, CoordinateSystem]:
    """Read all NXtransformations coordinate systems from the nexus file."""

    def transform(
        vector: NDArray[np.float64], matrix: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        return (matrix @ vector.T)[:-1]

    transformation_matrices = transformation_matrices_from(fname, include_process)

    coordinate_systems = {}

    for name, transformation_matrix in transformation_matrices.items():
        transform_vec = partial(transform, matrix=transformation_matrix)
        coordinate_systems[name] = CoordinateSystem(
            origin=transform_vec(np.array([0, 0, 0, 1])),
            x_axis=transform_vec(np.array([1, 0, 0, 0])),
            y_axis=transform_vec(np.array([0, 1, 0, 0])),
            z_axis=transform_vec(np.array([0, 0, 1, 0])),
        )

    return coordinate_systems


def unit_vector(vec: NDArray[np.float64]):
    """Calculates an unit vector"""
    return vec / np.linalg.norm(vec)


def angle_between(vec1: NDArray[np.float64], vec2: NDArray[np.float64]):
    """Calculate the angle between two vectors"""
    uvec1 = unit_vector(vec1)
    uvec2 = unit_vector(vec2)
    return np.arccos(np.clip(np.dot(uvec1, uvec2), -1, 1))



"""Transformation matrices for nexus"""
from typing import Optional
import numpy as np
from numpy.linalg import norm
from numpy.typing import NDArray


def rotate(
    angle: float,
    axis: NDArray[np.float64],
    offset: Optional[NDArray[np.float64]] = None,
) -> NDArray[np.float64]:
    """Generates a 4D rotation matrix

    Returns:
        np.ndarray[(4, 4), float]: The 4D rotation matrix
    """
    if offset is None:
        offset = np.zeros(3)

    v_x, v_y, v_z = axis / norm(axis)
    cosa = np.cos(angle)
    cosa1 = 1 - cosa
    sina = np.sin(angle)

    rot_matrix = np.array(
        [
            [
                cosa + v_x**2 * cosa1,
                v_x * v_y * cosa1 - v_z * sina,
                v_x * v_z * cosa1 + v_y * sina,
                offset[0],
            ],
            [
                v_y * v_x * cosa1 + v_z * sina,
                cosa + v_y**2 * cosa1,
                v_y * v_z * cosa1 - v_x * sina,
                offset[1],
            ],
            [
                v_z * v_x * cosa1 - v_y * sina,
                v_z * v_y * cosa1 + v_x * sina,
                cosa + v_z**2 * cosa1,
                offset[2],
            ],
            [0, 0, 0, 1],
        ]
    )

    return rot_matrix


def rotate_z_onto_vec(
    vec: NDArray[np.float64], offset: Optional[NDArray[np.float64]] = None
) -> NDArray[np.float64]:
    """Generates a 4D rotation matrix to transform the z-axis
    from the reference frame onto the provided vector.

    Args:
        vec (NDArray[np.float64]): The vector to rotate onto.
        offset (Optional[NDArray[np.float64]], optional):
            An additional offset. Setting this to None ignores the offset. Defaults to None.

    Returns:
        np.ndarray[(4, 4), float]: The 4D rotation matrix
    """
    if offset is None:
        offset = np.zeros(3)

    return np.array(
        [
            [0, 0, vec[0], offset[0]],
            [0, 0, vec[1], offset[1]],
            [0, 0, vec[2], offset[2]],
            [0, 0, 0, 1],
        ]
    )


def translate(
    translation: NDArray[np.float64], offset: Optional[NDArray[np.float64]] = None
) -> NDArray[np.float64]:
    """Generates a 4D translation matrix.

    Returns:
        np.ndarray[(4, 4), float]: The translation matrix.
    """
    trans = translation
    if offset is not None:
        trans += offset

    return np.array(
        [
            [1, 0, 0, trans[0]],
            [0, 1, 0, trans[1]],
            [0, 0, 1, trans[2]],
            [0, 0, 0, 1],
        ]
    )



"""Tests the CLI"""
import os
from click.testing import CliRunner
import pytest

from nexus3d.nexus_transformations import cli


@pytest.mark.filterwarnings("ignore: This file .* contains a binary blob")
def test_cli(tmp_path):
    """A test for the convert CLI."""
    path = os.getenv("PYTEST_CURRENT_TEST").rsplit("/", 1)[0]
    output_files = [
        tmp_path / f"experiment.{file_ending}" for file_ending in ["stl", "glb", "gltf"]
    ]

    for output_file in output_files:
        cli_input = [
            f"{path}/data/transformation_example.h5",
            "--output",
            str(output_file),
        ]
        runner = CliRunner()
        result = runner.invoke(cli, cli_input)

        assert result.exit_code == 0



"""Functions for creating a stl cube mesh file"""
from typing import Dict, Mapping, Union
import numpy as np
from numpy.typing import NDArray
from stl import mesh

from nexus3d.formats.mesh import create_cube_arrays
from nexus3d.formats.interfaces import WriterInput

TransformationMatrixDict = Mapping[
    str, Union[Dict[str, NDArray[np.float64]], NDArray[np.float64]]
]


def create_cube_mesh(scale: float = 1.0):
    """Creates a cube mesh.

    Args:
        scale (float, optional): The scale of the cube mesh. Defaults to 1.0.

    Returns:
        stl.mesh: The cube mesh at the origin
    """
    indices, vertices = create_cube_arrays(scale)

    cube = mesh.Mesh(np.zeros(indices.shape[0], dtype=mesh.Mesh.dtype))
    for i, face in enumerate(indices):
        for j in range(3):
            cube.vectors[i][j] = vertices[face[j], :]

    return cube


def cube_meshs_from(
    transformation_matrices: TransformationMatrixDict, scale: float = 0.1
) -> mesh.Mesh:
    """Creates a composed cube mesh for a dict of transformation matrices.

    Args:
        transformation_matrices (Dict[str, np.ndarray[): The transformation matrix dict.
        scale (float): The scale of the cubes. Defaults to 0.1.

    Returns:
        mesh: The composed mesh containing a cube for each transformation matrix.
    """
    scene = None
    for transformation_matrix in transformation_matrices.values():
        cube = create_cube_mesh(scale)
        cube.transform(transformation_matrix)

        if scene is None:
            scene = cube.data
            continue
        scene = np.concatenate((scene, cube.data))

    return mesh.Mesh(scene)


def write_stl_file(cli_input: WriterInput):
    """Writes a cube mesh from the transformation matrices to a stl file.

    Args:
        filename (str): The filename to write to.
        transformation_matrices (Dict[str, NDArray[np.float64]]): The transformation matrix dict.
        scale (float, optional): The scale of the cubes. Defaults to 0.1.
    """
    if isinstance(next(iter(cli_input.transformation_matrices.values())), dict):
        raise NotImplementedError(
            "Storage of intermediate matrices is not supported for stl files."
        )

    scene = cube_meshs_from(cli_input.transformation_matrices, cli_input.size / 2)
    scene.save(cli_input.output)



"""Test correct reading and reduction of transformations"""
import os
import numpy as np
from numpy.testing import assert_array_almost_equal, assert_almost_equal
from pytest import fixture
from scipy.spatial.transform import Rotation
from nexus3d.coordinate_systems import angle_between

from nexus3d.matrix import rotate, translate
from nexus3d.nexus_transformations import transformation_matrices_from

# pylint: disable=redefined-outer-name


@fixture
def example_file_path():
    """Get the transformation example file path."""
    path = os.getenv("PYTEST_CURRENT_TEST").rsplit("/", 1)[0]
    return f"{path}/data/transformation_example.h5"


@fixture
def tmatrices(example_file_path):
    """Get the transformation matrixes from the example file."""
    return transformation_matrices_from(example_file_path, False)


def test_correct_rotation_matrix():
    """Checks the nexus rotation matrix"""
    angles = np.linspace(0, 360, 36)
    rotvecs = np.random.rand(36, 3)

    for angle, rotvec in zip(angles, rotvecs):
        rot_matrix = np.pad(
            Rotation.from_rotvec(angle * rotvec / np.linalg.norm(rotvec)).as_matrix(),
            ((0, 1), (0, 1)),
        )
        rot_matrix[-1, -1] = 1
        nexus_matrix = rotate(angle, rotvec, np.array([0, 0, 0]))

        assert_array_almost_equal(nexus_matrix, rot_matrix)


def test_correct_chain_resolution_from_nexus(tmatrices):
    """The transformation chain is resolved to the correct matrix from a nexus file"""

    def get_manipulator_trafo():
        m_rot_x = rotate(np.deg2rad(-90), np.array([1, 0, 0]))
        m_rot_z = rotate(np.deg2rad(-25), np.array([0, 0, 1]))
        m_trans_z = translate(-0.32 * np.array([0, 0, 1]))

        return m_rot_x @ m_rot_z @ m_trans_z

    def get_sample_trafo():
        trans_x = translate(2.65e-3 * np.array([1, 0, 0]))
        trans_y = translate(-4.321e-3 * np.array([0, 1, 0]))
        trans_z = translate(34.82e-3 * np.array([0, 0, 1]))
        rot_tht = rotate(np.deg2rad(401.18), np.array([0, 0, 1]))
        rot_phi = rotate(np.deg2rad(2.5), np.array([0, 1, 0]))
        rot_omg = rotate(np.deg2rad(131.7), np.array([1, 0, 0]))
        corrected_phi = rotate(np.deg2rad(90), np.array([0, 1, 0]))

        return (
            get_manipulator_trafo()
            @ trans_x
            @ trans_y
            @ trans_z
            @ rot_tht
            @ rot_phi
            @ rot_omg
            @ corrected_phi
        )

    def get_analyser_trafo():
        a_rot_y = rotate(np.deg2rad(-115), np.array([0, 1, 0]))
        a_trans_z = translate(4e-3 * np.array([0, 0, 1]))

        return a_rot_y @ a_trans_z

    assert_array_almost_equal(
        tmatrices["instrument/manipulator"], get_manipulator_trafo()
    )
    assert_array_almost_equal(tmatrices["sample"], get_sample_trafo())

    assert_array_almost_equal(
        tmatrices["instrument/electronanalyser"], get_analyser_trafo()
    )


def test_full_chain_extraction(tmatrices, example_file_path):
    """Test the full chain extraction from the example file."""
    tmatrices_chain = transformation_matrices_from(example_file_path, False, True)

    assert len(tmatrices_chain["sample"]) == 10
    assert len(tmatrices_chain["instrument/manipulator"]) == 3
    assert len(tmatrices_chain["instrument/electronanalyser"]) == 2

    for entry, matrix in tmatrices.items():
        last_matrix = tmatrices_chain[entry][next(reversed(tmatrices_chain[entry]))]
        assert_array_almost_equal(matrix, last_matrix)


def test_angle_between(tmatrices):
    """Test if the angle between the sample z-axis and beam axis is calculated correctly"""
    transformed_z = (
        np.linalg.inv(tmatrices["instrument/electronanalyser"]) @ np.array([0, 0, 1, 0])
    )[:-1]

    assert_almost_equal(
        angle_between(transformed_z, np.array([0, 0, -1])) / np.pi * 180,
        65,
    )



"""Create a stl from NXtransformation groups"""
from collections import OrderedDict
import json
import os
from sys import version_info
from os import path
from typing import Callable, Dict, Mapping, Union
import numpy as np
from numpy.typing import NDArray
import h5py
import click
from nexus3d.formats.interfaces import WriterInput

from nexus3d.matrix import rotate, translate
from nexus3d.formats.stl_writer import write_stl_file
from nexus3d.formats.gltf_writer import write_gltf_file
from nexus3d.units import ureg

TransformationMatrixDict = Mapping[
    str, Union[Dict[str, NDArray[np.float64]], NDArray[np.float64]]
]


def transformation_matrices_from(
    fname: str, include_process: bool, store_intermediate: bool = False
) -> TransformationMatrixDict:
    """Reads all NXtransformations from a nexus file
    and creates a transformation matrix from them."""

    def store_in_chain(entry: str, matrix: NDArray[np.float64]):
        if store_intermediate:
            matrix_chain[entry] = matrix

        return matrix

    def get_transformation_matrix(h5file: h5py.File, entry: str):
        required_attrs = ["depends_on", "vector", "transformation_type", "units"]
        attrs = h5file[entry].attrs

        for req_attr in required_attrs:
            if req_attr not in attrs:
                raise ValueError(f"`{req_attr}` attribute not found in {entry}")

        if "offset" in attrs and "offsets_units" not in attrs:
            raise ValueError(
                f"Found `offset` attribute in {entry} but no `offset_units` could be found."
            )
        offset = attrs.get("offset", np.zeros((3,)))
        offset_unit = ureg(attrs.get("offset_unit", ""))
        offset_si = (offset * offset_unit).to_base_units().magnitude

        vector = attrs["vector"]

        # Always use the first element and ignore the dimension
        # With gltf we probably could also animate it!
        field = h5file[entry][()].flat[0]
        if not isinstance(field, (int, float, np.int64, np.float64)):
            raise NotImplementedError(
                "Only float fields are supported yet, "
                f"but {entry}: {field} is {type(field)}"
            )
        field_si = ureg(f"{field} {attrs['units']}").to_base_units().magnitude  # type: ignore

        if attrs["transformation_type"] == "translation":
            matrix = translate(field_si * vector, offset_si)
        elif attrs["transformation_type"] == "rotation":
            matrix = rotate(field_si, vector, offset_si)
        else:
            raise ValueError(
                f"Unknown transformation type `{attrs['transformation_type']}`"
            )

        if attrs["depends_on"] == ".":
            return store_in_chain(entry, matrix)

        if "/" in attrs["depends_on"]:
            return store_in_chain(
                entry, get_transformation_matrix(h5file, attrs["depends_on"]) @ matrix
            )

        return store_in_chain(
            entry,
            get_transformation_matrix(
                h5file, f"{entry.rsplit('/', 1)[0]}/{attrs['depends_on']}"
            )
            @ matrix,
        )

    def get_transformation_group_names(name: str, dataset: h5py.Dataset):
        if not include_process and name.startswith("entry/process"):
            return

        if "depends_on" in name:
            transformation_groups[
                name.rsplit("/", 1)[0].split("/", 1)[-1]
                if version_info.minor < 9
                else name.removesuffix("/depends_on").removeprefix("entry/")
            ] = dataset[()].decode("utf-8")

    transformation_groups: Dict[str, h5py.Dataset] = {}
    transformation_matrices = {}
    with h5py.File(fname, "r") as h5file:
        h5file.visititems(get_transformation_group_names)

        for name, transformation_group in transformation_groups.items():
            if store_intermediate:
                matrix_chain: Dict[str, NDArray[np.float64]] = OrderedDict()

            transformation_matrix = get_transformation_matrix(
                h5file, transformation_group
            )

            transformation_matrices[name] = (
                matrix_chain if store_intermediate else transformation_matrix
            )

    return transformation_matrices


@click.command()
@click.argument("file")
@click.option(
    "-o",
    "--output",
    default="experiment.glb",
    help="The filename to write to.",
    show_default=True,
)
@click.option(
    "-c",
    "--config",
    default="",
    help="Config file to load stl drawings into the final output",
)
@click.option(
    "-s",
    "--size",
    default=0.1,
    type=float,
    help="The side length of a cube in meters.",
    show_default=True,
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    type=bool,
    help="Force overwriting of output file.",
)
@click.option(
    "--include-process",
    is_flag=True,
    default=False,
    type=bool,
    help="Include transformations inside /entry/process",
)
@click.option(
    "--store-intermediate",
    is_flag=True,
    default=False,
    type=bool,
    help=(
        "Store the intermediate matrices in gltf child nodes. "
        "Only applicable for gltf or glb files."
    ),
)
@click.option(
    "--shape",
    default="cone",
    type=click.Choice(["cone", "cube"]),
    help=(
        "The shape to write into the gltf file. Only applicable for gltf or glb files. "
    ),
    show_default=True,
)
def cli(  # pylint: disable=too-many-arguments
    file: str,
    config: str,
    output: str,
    force: bool,
    size: float,
    include_process: bool,
    store_intermediate: bool,
    shape: str,
):
    """
    Create a glb/gltf or stl from a nexus file via the command line.
    The actual file format is chosen from the
    file ending of the output file (default: experiment.glb).
    """

    if not path.exists(file):
        raise click.FileError(file, hint=f"File `{file}` does not exist.")

    if not path.isfile(file):
        raise click.FileError(file, hint=f"`{file}` is not a file.")

    if not h5py.is_hdf5(file):
        raise click.FileError(file, hint=f"`{file}` is not a valid HDF5 file.")

    if path.exists(output) and not force:
        raise click.FileError(
            output, hint=f"File `{output}` already exists. Use -f to overwrite."
        )

    file_format = os.path.splitext(output)[1]
    if file_format not in [".stl", ".gltf", ".glb"]:
        raise click.UsageError(
            message=f"Unsupported file format {file_format} in output file `{output}`"
        )

    if config and (not path.exists(config) or not path.isfile(config)):
        raise click.FileError(
            file, hint=f"Config file `{config}` does not exist or is not valid."
        )

    if config and os.path.splitext(config)[1] != ".json":
        raise click.FileError(file, hint=f"Config file `{config}` must be a json file.")

    if size <= 0:
        raise click.BadOptionUsage(
            "size", f"Not a valid size: {size}. Size needs to be > 0."
        )

    def format_not_implemented(cli_input: WriterInput):
        raise NotImplementedError(
            f"Cannot write to format {file_format} for output file {cli_input.output}"
        )

    config_dict = {}
    if config:
        with open(config, "r", encoding="utf-8") as json_file:
            config_dict = json.load(json_file)

    format_map: Dict[str, Callable[[WriterInput], None]] = {
        ".stl": write_stl_file,
        ".gltf": write_gltf_file,
        ".glb": write_gltf_file,
    }

    format_map.get(file_format, format_not_implemented)(
        WriterInput(
            output=output,
            transformation_matrices=transformation_matrices_from(
                file, include_process, store_intermediate
            ),
            size=size,
            show_beam=True,
            config_dict=config_dict,
            shape=shape,
        )
    )



"""Cube mesh utility functions (e.g. creating cube arrays)"""
from typing import Optional
import numpy as np
from stl import mesh

from nexus3d.units import ureg


def create_cube_arrays(scale: float = 1):
    """Get vertices and indices arrays for creating a cube.

    Args:
        scale (float, optional): The scale of the cube. Defaults to 1.

    Returns:
        (np.ndarray, np.ndarray): The points and triangles array of the cube.
    """
    vertices = np.array(
        [
            [-0.5, -1, -2],
            [+0.5, -1, -2],
            [+0.5, +1, -2],
            [-0.5, +1, -2],
            [-0.5, -1, +2],
            [+0.5, -1, +2],
            [+0.5, +1, +2],
            [-0.5, +1, +2],
        ],
        dtype="float32",
    )
    indices = np.array(
        [
            [0, 3, 1],
            [1, 3, 2],
            [0, 4, 7],
            [0, 7, 3],
            [4, 5, 6],
            [4, 6, 7],
            [5, 1, 2],
            [5, 2, 6],
            [2, 3, 6],
            [3, 7, 6],
            [0, 1, 5],
            [0, 5, 4],
        ],
        dtype="uint8",
    )

    return indices, vertices * scale


def create_cone_arrays(scale: float = 1):
    """Get vertices and index arrays for creating a cone.

    Args:
        scale (float, optional): The scale of the cone. Defaults to 1.

    Returns:
        (np.ndarray, np.ndarray): The points and triangles array of the cone.
    """
    aspect_ratio = 2
    vertices = (
        np.array(
            [
                [0, 0, 0],
                [-1.3, 0, -aspect_ratio],
                [-0.809017, -0.587785, -aspect_ratio],
                [-0.309017, -0.951057, -aspect_ratio],
                [0.309017, -0.951057, -aspect_ratio],
                [0.809017, -0.587785, -aspect_ratio],
                [1.3, 0, -aspect_ratio],
                [0.809017, 0.587785, -aspect_ratio],
                [0.309017, 0.951057, -aspect_ratio],
                [-0.309017, 0.951057, -aspect_ratio],
                [-0.809017, 0.587785, -aspect_ratio],
            ],
            dtype="float32",
        )
        / aspect_ratio
    )

    indices = np.array(
        [
            [0, 1, 2],
            [0, 2, 3],
            [0, 3, 4],
            [0, 4, 5],
            [0, 5, 6],
            [0, 6, 7],
            [0, 7, 8],
            [0, 8, 9],
            [0, 9, 10],
            [0, 10, 1],
        ],
        dtype="uint8",
    )

    return indices, vertices * scale


def get_mesh_from_stl(filename: str, unit: Optional[str] = None):
    """Reads a mesh as array of indices and vertices from a stl file.

    Args:
        filename (str): The stl filename
        unit (str): A pint interpretable unit which to interpret from the stl file.
    """
    stl_mesh = mesh.Mesh.from_file(filename)

    vertices, indices_lin = np.unique(
        stl_mesh.vectors.reshape([stl_mesh.vectors.size // 3, 3]),
        axis=0,
        return_inverse=True,
    )

    for dtype in ["uint8", "uint16", "uint32"]:
        if indices_lin.max() < np.iinfo(dtype).max:
            break

        if dtype == "uint32":
            raise ValueError(
                f"The stl model `{filename}` is too large to be converted into gltf."
            )

    indices = indices_lin.reshape([indices_lin.size // 3, 3]).astype(dtype)

    if unit is not None:
        scaling = ureg(f"1 {unit}").to("m").magnitude  # type: ignore
        return indices, vertices * scaling
    return indices, vertices



"""Tests for stl file functions"""
import numpy as np
from numpy.testing import assert_array_almost_equal
from nexus3d.formats.mesh import create_cube_arrays, get_mesh_from_stl
from nexus3d.formats.interfaces import WriterInput
from nexus3d.formats.stl_writer import write_stl_file


def test_cube_mesh_reconstruction(tmp_path):
    """Test whether a cube mesh is correctly reconstructed from a stl file"""
    test_file = tmp_path / "cube.stl"
    write_stl_file(WriterInput(test_file, {"test": np.identity(4)}, 2.0, False, {}))

    indices, vertices = create_cube_arrays()
    indices_file, vertices_file = get_mesh_from_stl(test_file)

    for i, face in enumerate(indices):
        for j in range(3):
            assert_array_almost_equal(
                vertices[face[j], :], vertices_file[indices_file[i][j], :]
            )
